# Copyright 2024 Polyquantique

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
haarpy Python interface
"""

from math import factorial
from copy import deepcopy
from itertools import product
from fractions import Fraction
import numpy as np
from sympy import Symbol, fraction, simplify, factor
from sympy.combinatorics import Permutation
from sympy.utilities.iterables import partitions


def get_class(cycle: Permutation, degree: int) -> tuple:
    """Returns the conjugacy class of an element of the symmetric group Sp

    Args:
        degree (integer): Order of the symmetric group
        cycle (cycle): Permutation cycle from the symmetric group

    Returns:
        tuple[int]: the conjugacy class in partition form

    Raise:
        TypeError : order must be of type int
        ValueError : If order is not an integer greaten than 1
        TypeError : cycle must be of type sympy.combinatorics.permutations.Permutation
        ValueError : Incompatible degree and permutation cycle
    """
    if not isinstance(degree, int):
        raise TypeError("degree must be of type int")
    if degree < 2:
        raise ValueError(
            "The degree you have provided is too low. It must be an integer greater than 1."
        )
    if not isinstance(cycle, Permutation):
        raise TypeError(
            "cycle must be of type sympy.combinatorics.permutations.Permutation"
        )

    if cycle.support() == []:
        return degree * (1,)

    if max(cycle.support()) >= degree:
        raise ValueError("Incompatible degree and permutation cycle")
    return tuple(sorted([len(i) for i in cycle.cyclic_form], reverse=True)) + (1,) * (
        degree - len(cycle.support())
    )


def derivative_tableau(
    tableau: list[list[int]], add_unit: int, partition: tuple[int]
) -> list[list[list[int]]]:
    """Takes a single tableau and adds the selected number to it's contents
    in a way that keeps it semi-standard

    Args:
        tableau (list[list[int]]): The tableau
        add_unit (int): Selected number to be added
        partition ([list[int]) : partition characterizing an irrep of Sp

    Returns:
        list[list[list[int]]]: Modified tableau
    """
    tableau_derivatives = []
    for i, _ in enumerate(partition):
        for j in range(partition[i]):
            if tableau[i][j] == 0:
                cond = i == 0 and j == 0 and add_unit == 1
                cond |= (
                    j == 0 and tableau[i - 1][0] <= add_unit and tableau[i - 1][0] != 0
                )
                cond |= (
                    i == 0 and tableau[0][j - 1] <= add_unit and tableau[0][j - 1] != 0
                )
                cond |= (
                    i != 0
                    and j != 0
                    and tableau[i - 1][j] <= add_unit
                    and tableau[i][j - 1] <= add_unit
                    and tableau[i - 1][j] != 0
                    and tableau[i][j - 1] != 0
                )
                if cond:
                    copy = deepcopy(tableau)
                    copy[i][j] = add_unit
                    tableau_derivatives.append(copy)
    return tableau_derivatives


def ssyt(partition: tuple[int], conjugacy_class: tuple[int]) -> list[list[int]]:
    """all eligible semi-standard young tableaux based of the partition

    Args:
        partition ([list[int]) : partition characterizing an irrep of Sp
        conjugacy_class (list[int]) : A conjugacy class, in partition form, of Sp

    Returns:
        list[list[list[int]]]: all eligible semi-standard young tableaux
    """
    tableau = [[partition[0] * [0] for _ in range(len(partition))]]
    add_unit = [i + 1 for i, m in enumerate(conjugacy_class) for j in range(m)]

    for unit in add_unit:
        tableau = [derivative_tableau(see, unit, partition) for see in tableau]
        tableau = [i for j in tableau for i in j]  # flattens the list
        tableau = {tuple(tuple(i) for i in j) for j in tableau}  # removes duplicates
        tableau = [[list(i) for i in j] for j in tableau]

    return tableau


def bad_mapping(tableau: list[list[int]], conjugacy_class: tuple[int]) -> bool:
    """Flags tableaux that have a bad mapping

    Args:
        tableau (list[list[int]]) : A semi-standard Young tableau
        conjugacy_class (list[int]) : A conjugacy class, in partition form, of Sp

    Returns:
        bool : True if mapping is wrong
    """
    for i in range(len(conjugacy_class)):
        # indices of matching elements
        matching = [
            (j + 1, k + 1)
            for j, tab in enumerate(tableau)
            for k, val in enumerate(tab)
            if val == i + 1
        ]
        counting = len(matching)
        if counting == 1:
            matching.append((matching[0][0] + 1, matching[0][1]))

        shape = (counting, counting, 2, 2)
        matching = np.array(list(product(matching, repeat=2)))
        matching = matching.flatten()[: np.prod(shape)]
        matching = matching.reshape(shape).tolist()

        # Euclidean distance
        for j in range(counting):
            for k in range(counting):
                matching[j][k] = np.linalg.norm(
                    np.array(matching[j][k][1]) - np.array(matching[j][k][0])
                )

        d = [matching[j].count(1) for j in range(counting)]

        for val in d:
            if counting >= 2 and val == 0:
                return True

        if d.count(1) > 2 or (d.count(1) == 0 and counting != 1):
            return True

    return False


def murn_naka(partition: tuple[int], conjugacy_class: tuple[int]) -> int:
    """Implementation of the Murnaghan-Nakayama rule for the characters irreducible
    representations of the symmetric group Sp

    Args:
        partition (tuple[int]) : partition characterizing an irrep of Sp
        conjugacy_class (tuple[int]) : A conjugacy class, in partition form, of Sp

    Returns:
        int : Character of the elements in class mu of the irrep of the symmetric group
    """
    if sum(partition) != sum(conjugacy_class):
        return 0

    tableaux_list = ssyt(partition, conjugacy_class)
    tableaux_list = [
        tableau
        for tableau in tableaux_list
        if not bad_mapping(tableau, conjugacy_class)
    ]

    skip = False
    tableaux_list = np.array(tableaux_list)
    for i, tableau in enumerate(tableaux_list):
        for j in range(len(partition) - 1):
            if skip:
                skip = False
                break
            for k in range(partition[0] - 1):
                if (
                    tableau[j, k]
                    and tableau[j, k]
                    == tableau[j, k + 1]
                    == tableau[j + 1, k]
                    == tableau[j + 1, k + 1]
                ):
                    np.delete(tableaux_list, i, 0)
                    skip = True
                    break

    height = np.empty(shape=(len(tableaux_list), len(conjugacy_class)))
    height.fill(-1)
    weight = np.empty(shape=(len(tableaux_list),))
    for i, tableau in enumerate(tableaux_list):
        for j in range(len(conjugacy_class)):
            for k in range(len(partition)):
                height[i, j] += 1 if np.count_nonzero(tableau[k] == j + 1) else 0
        weight[i] = (-1) ** np.sum(height[i, : len(conjugacy_class)])

    return int(np.sum(weight[: len(tableaux_list)]))


def sn_dimension(partition: tuple[int]) -> int:
    """Returns the dimension of the irrep of the symmetric group Sp labelled by the input partition

    Args:
        partition (list[int]) : A partition labelling an irrep of Sp

    Returns:
        int : The dimension of the irrep
    """
    numerator = np.prod(
        [
            part_i - part_j + j + 1
            for i, part_i in enumerate(partition)
            for j, part_j in enumerate(partition[i + 1 :])
        ],
        dtype=np.int64,
    )
    len_partition = len(partition)
    denominator = np.prod(
        [factorial(part + len_partition - i - 1) for i, part in enumerate(partition)],
        dtype=np.int64,
    )
    dimension = Fraction(numerator, denominator) * factorial(sum(partition))
    if dimension.denominator != 1:
        raise ValueError("Fraction too large for dtype int32")

    return dimension.numerator


def ud_dimension(partition: tuple[int], unitary_dimension: Symbol) -> int:
    """Returns the dimension of the unitary group U(d) labelled by the input partition

    Args:
        partition (list[int]) : A partition labelling U(d)
        unitary_dimension (Symbol) : dimension d of the unitary matrix U

    Returns:
        Symbol : The dimension of U(d)
    """
    conjugate_partition = [
        sum(1 for _, part in enumerate(partition) if i < part)
        for i in range(partition[0])
    ]
    if isinstance(unitary_dimension, int):
        dimension = np.prod(
            [
                np.prod(
                    [
                        Fraction(
                            unitary_dimension + j - i,
                            part + conjugate_partition[j] - i - j - 1,
                        )
                        for j in range(part)
                    ]
                )
                for i, part in enumerate(partition)
            ]
        )
        if dimension.denominator != 1:
            raise ValueError("Fraction too large for dtype int32")
        return dimension.numerator

    numerator = np.prod(
        [
            np.prod([(unitary_dimension + k - i) for k in range(part)])
            for i, part in enumerate(partition)
        ]
    )

    denominator = np.prod(
        [
            np.prod([part + conjugate_partition[j] - i - j - 1 for j in range(part)])
            for i, part in enumerate(partition)
        ]
    )

    return numerator / denominator


def weingarten_class(conjugacy_class: tuple[int], unitary_dimension: Symbol) -> Symbol:
    """Returns the Weingarten function

    Args:
        conjugacy_class (tuple[int]) : A conjugacy class, in partition form, of Sp
        unitary_dimension (Symbol) : Dimension d of the unitary matrix U

    Returns:
        Symbol : The Weingarten function

    Raise:
        TypeError : unitary_dimension must be an instance of int or sympy.Symbol
    """
    if not isinstance(unitary_dimension, (Symbol, int)):
        raise TypeError("unitary_dimension must be an instance of int or sympy.Symbol")

    degree = sum(conjugacy_class)
    partition_list = [
        sum([value * (key,) for key, value in part.items()], ())
        for part in partitions(degree)
    ]
    irrep_dimension_list = [sn_dimension(part) for part in partition_list]

    if isinstance(unitary_dimension, int):
        weingarten = sum(
            Fraction(
                irrep_dimension**2 * murn_naka(part, conjugacy_class),
                ud_dimension(part, unitary_dimension),
            )
            for part, irrep_dimension in zip(partition_list, irrep_dimension_list)
        ) * Fraction(1, factorial(degree) ** 2)
    else:
        weingarten = (
            sum(
                irrep_dimension**2
                * murn_naka(part, conjugacy_class)
                / ud_dimension(part, unitary_dimension)
                for part, irrep_dimension in zip(partition_list, irrep_dimension_list)
            )
            / factorial(degree) ** 2
        )
        numerator, denominator = fraction(simplify(weingarten))
        weingarten = factor(numerator) / factor(denominator)

    return weingarten


def weingarten_element(
    cycle: Permutation, degree: int, unitary_dimension: Symbol
) -> Symbol:
    """Returns the Weingarten function

    Args:
        cycle (Permutation) : Permutation cycle from the symmetric group
        degree (int) : Degree p of the symmetric group Sp
        unitary_dimension (Symbol) : Dimension d of the unitary matrix U

    Returns:
        Symbol : The Weingarten function
    """
    conjugacy_class = list(get_class(cycle, degree))
    return weingarten_class(conjugacy_class, unitary_dimension)



# Copyright 2024 Polyquantique

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Haarpy
==========

.. currentmodule:: Haarpy

This is the top level module of Haarpy, containing functions for
the symbolic calculation of Weingarten functions and related averages
of unitary matrices sampled uniformly at random from the Haar measure.

Functions
---------

.. autosummary::
    get_class
    derivative_tableau
    ssyt
    bad_mapping
    murn_naka
    sn_dimension
    ud_dimension
    weingarten_class
    weingarten_element

Code details
------------
"""
from .unitary import (
    get_class,
    derivative_tableau,
    ssyt,
    bad_mapping,
    murn_naka,
    sn_dimension,
    ud_dimension,
    weingarten_class,
    weingarten_element,
)

from ._version import __version__

__all__ = [
    "get_class",
    "derivative_tableau",
    "ssyt",
    "bad_mapping",
    "murn_naka",
    "sn_dimension",
    "ud_dimension",
    "weingarten_class",
    "weingarten_element",
]


def version():
    r"""
    Get version number of haarpy

    Returns:
      str: The package version number
    """
    return __version__


def about():
    """Haarpy information.

    Prints the installed version numbers for Haarpy and its dependencies,
    and some system info. Please include this information in bug reports.

    **Example:**

    .. code-block:: pycon

        >>> haarpy.about()
        Haarpy: a Python library for the symbolic calculation of Weingarten functions.

        Python version:            3.12.3
        Platform info:             Linux-6.8.0-31-generic-x86_64-with-glibc2.39
        Installation path:         /home/username/haarpy
        Haarpy version:            0.0.1
        Numpy version:             1.26.4
        SymPy version:             1.12
    """
    # pylint: disable=import-outside-toplevel
    import os
    import platform

    import sys
    import numpy
    import sympy

    # a QuTiP-style infobox
    print(
        "\nHaarpy: a Python library for the symbolic calculation of Weingarten functions."
    )
    # print("Copyright 2018-2021 Polyquantique\n")

    print("Python version:            {}.{}.{}".format(*sys.version_info[0:3]))
    print(f"Platform info:             {platform.platform()}")
    print(f"Installation path:         {os.path.dirname(__file__)}")
    print(f"Haarpy version:            {__version__}")
    print(f"Numpy version:             {numpy.__version__}")
    print(f"SymPy version:             {sympy.__version__}")



# Copyright 2024 Polyquantique

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages


with open("haarpy/_version.py") as f:
    version = f.readlines()[-1].split()[-1].strip("\"'")


requirements = [
    "numpy>=1.26.4",
    "sympy>=1.12"
]

classifiers = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: Apache Software License",
    "Natural Language :: English",
    "Operating System :: POSIX",
    "Operating System :: MacOS :: MacOS X",
    "Operating System :: POSIX :: Linux",
    "Operating System :: Microsoft :: Windows",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
    "Programming Language :: Python :: 3 :: Only",
    "Topic :: Scientific/Engineering :: Physics",
]


setup(
    name="haarpy",
    version=version,
    description="Symbolic calculation of Weingarten functions",
    url="https://github.com/polyquantique/haarpy",
    author="Yanic Cardin",
    author_email="yanic.cardin@polymtl.ca",
    license="Apache License 2.0",
    packages=find_packages(where="."),
    install_requires=requirements,
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    classifiers=classifiers,
)



# Copyright 2024 Polyquantique

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
haarpy tests
"""

import pytest
import haarpy as ap
from fractions import Fraction
from sympy.combinatorics import Permutation
from numpy import allclose
from sympy import Symbol


@pytest.mark.parametrize(
    "degree,cycle,conjugacy",
    [
        (3, Permutation(0, 1, 2), (3,)),
        (
            3,
            Permutation(
                2,
            )(0, 1),
            (2, 1),
        ),
        (4, Permutation(0, 1, 2, 3), (4,)),
        (4, Permutation(0, 1, 2), (3, 1)),
        (4, Permutation(0, 1)(2, 3), (2, 2)),
        (5, Permutation(0, 1, 2), (3, 1, 1)),
        (5, Permutation(1, 2, 4), (3, 1, 1)),
        (
            5,
            Permutation(
                3,
            )(0, 1, 2),
            (3, 1, 1),
        ),
        (
            6,
            Permutation(
                2,
            )(
                3, 4, 5
            )(0, 1),
            (3, 2, 1),
        ),
        (6, Permutation(2, 3, 4, 0, 1, 5), (6,)),
        (7, Permutation(2, 3, 4, 0, 1, 6), (6, 1)),
    ],
)
def test_get_class(degree, cycle, conjugacy):
    """Test the normal usage of get_class"""
    assert ap.get_class(cycle, degree) == conjugacy


@pytest.mark.parametrize(
    "degree,cycle",
    [
        ("a", Permutation(0, 1, 2)),
        (0.1, Permutation(2)(0, 1)),
        ((1,), Permutation(0, 1, 2, 3)),
        ([5], Permutation(0, 1, 2)),
    ],
)
def test_get_class_degree_type_error(degree, cycle):
    """Test the degree parameter TypeError"""
    with pytest.raises(TypeError, match=".*degree must be of type int.*"):
        ap.get_class(cycle, degree)


@pytest.mark.parametrize("degree", range(-2, 2))
def test_get_class_degree_value_error(degree):
    """Test the degree parameter ValueError"""
    with pytest.raises(
        ValueError,
        match=".*The degree you have provided is too low. It must be an integer greater than 1.*",
    ):
        ap.get_class(Permutation(0, 1, 2), degree)


@pytest.mark.parametrize(
    "degree,cycle",
    [
        (3, [1, 2, 3]),
        (4, "a"),
        (7, 2.0),
    ],
)
def test_get_class_cycle_type_error(degree, cycle):
    """Test the cycle parameter TypeError"""
    with pytest.raises(
        TypeError,
        match=".*cycle must be of type sympy.combinatorics.permutations.Permutation.*",
    ):
        ap.get_class(cycle, degree)


@pytest.mark.parametrize(
    "degree,cycle",
    [
        (3, Permutation(0, 1, 2, 3)),
        (3, Permutation(2, 3)(0, 1)),
        (4, Permutation(0, 1, 2, 3, 5)),
        (4, Permutation(4, 1)),
    ],
)
def test_get_class_cycle_value_error(degree, cycle):
    """Test the cycle parameter ValueError if permutation maximum value is greater than the degree"""
    with pytest.raises(
        ValueError, match=".*Incompatible degree and permutation cycle.*"
    ):
        ap.get_class(cycle, degree)


@pytest.mark.parametrize(
    "tableau,add_unit,partition,result",
    [
        ([[0, 0, 0], [0, 0, 0]], 1, (3, 1), [[[1, 0, 0], [0, 0, 0]]]),
        (
            [[1, 0, 0], [0, 0, 0]],
            1,
            (3, 1),
            [[[1, 1, 0], [0, 0, 0]], [[1, 0, 0], [1, 0, 0]]],
        ),
        (
            [[1, 2, 0], [0, 0, 0]],
            2,
            (3, 2),
            [[[1, 2, 2], [0, 0, 0]], [[1, 2, 0], [2, 0, 0]]],
        ),
        ([[1, 2, 0], [2, 3, 0]], 4, [3, 2], [[[1, 2, 4], [2, 3, 0]]]),
        (
            [[0, 0, 0], [0, 0, 0], [0, 0, 0]],
            1,
            (3, 1, 1),
            [[[1, 0, 0], [0, 0, 0], [0, 0, 0]]],
        ),
        ([[1, 1, 1, 3], [2, 0, 0, 0]], 4, (4, 2), [[[1, 1, 1, 3], [2, 4, 0, 0]]]),
        (
            [[1, 2, 3, 0], [1, 2, 0, 0], [1, 0, 0, 0]],
            3,
            (4, 2, 1),
            [[[1, 2, 3, 3], [1, 2, 0, 0], [1, 0, 0, 0]]],
        ),
    ],
)
def test_derivative_tableau(tableau, add_unit, partition, result):
    """Test the normal usage of derivative_tableau"""
    assert ap.derivative_tableau(tableau, add_unit, partition) == result


@pytest.mark.parametrize(
    "partition, conjugacy_class, result",
    [
        ((3, 1), (2, 2), [[[1, 2, 2], [1, 0, 0]], [[1, 1, 2], [2, 0, 0]]]),
        (
            (3, 1, 1),
            (3, 2),
            [
                [[1, 1, 1], [2, 0, 0], [2, 0, 0]],
                [[1, 1, 2], [1, 0, 0], [2, 0, 0]],
                [[1, 2, 2], [1, 0, 0], [1, 0, 0]],
            ],
        ),
        (
            (3, 2),
            (1, 2, 1, 1),
            [[[1, 2, 2], [3, 4, 0]], [[1, 2, 3], [2, 4, 0]], [[1, 2, 4], [2, 3, 0]]],
        ),
        (
            (4, 2),
            (2, 2, 2),
            [
                [[1, 1, 2, 3], [2, 3, 0, 0]],
                [[1, 2, 2, 3], [1, 3, 0, 0]],
                [[1, 2, 3, 3], [1, 2, 0, 0]],
                [[1, 1, 2, 2], [3, 3, 0, 0]],
                [[1, 1, 3, 3], [2, 2, 0, 0]],
            ],
        ),
        (
            (4, 3),
            (2, 2, 2),
            [
                [[1, 1, 2, 0], [2, 3, 3, 0]],
                [[1, 1, 2, 3], [2, 3, 0, 0]],
                [[1, 1, 3, 0], [2, 2, 3, 0]],
                [[1, 2, 2, 0], [1, 3, 3, 0]],
                [[1, 2, 3, 0], [1, 2, 3, 0]],
                [[1, 2, 2, 3], [1, 3, 0, 0]],
                [[1, 2, 3, 3], [1, 2, 0, 0]],
                [[1, 1, 2, 2], [3, 3, 0, 0]],
                [[1, 1, 3, 3], [2, 2, 0, 0]],
            ],
        ),
        ((4, 2), (2, 2, 2, 2), []),
    ],
)
def test_ssyt(partition, conjugacy_class, result):
    """Test the normal usage of ssyt"""
    assert ap.ssyt(partition, conjugacy_class) == result


@pytest.mark.parametrize(
    "tableau, conjugacy_class",
    [
        ([[1, 2, 2], [1, 0, 0]], (2, 2)),
        ([[1, 1, 1], [2, 0, 0]], (3, 1)),
        ([[1, 1, 2], [1, 0, 0]], (3, 1)),
        ([[1, 1, 1], [2, 0, 0], [2, 0, 0]], (3, 2)),
        ([[1, 2, 2], [1, 0, 0], [1, 0, 0]], (3, 2)),
        ([[1, 2, 2], [3, 4, 0]], (1, 2, 1, 1)),
        ([[1, 1, 2, 2], [3, 3, 0, 0]], (2, 2, 2)),
        ([[1, 1, 3, 3], [2, 2, 0, 0]], (2, 2, 2)),
        ([[1, 2, 3, 3], [1, 2, 0, 0]], (2, 2, 2)),
    ],
)
def test_bad_mapping_false(tableau, conjugacy_class):
    "Test bad_mapping for well mapped tableaux, returns False"
    assert not ap.bad_mapping(tableau, conjugacy_class)


@pytest.mark.parametrize(
    "tableau, conjugacy_class",
    [
        ([[1, 1, 2], [2, 0, 0]], (2, 2)),
        ([[1, 1, 2], [1, 0, 0], [2, 0, 0]], (3, 2)),
        ([[1, 2, 3], [2, 4, 0]], (1, 2, 1, 1)),
        ([[1, 2, 4], [2, 3, 0]], (1, 2, 1, 1)),
        ([[1, 1, 2, 3], [2, 3, 0, 0]], (2, 2, 2)),
        ([[1, 2, 2, 3], [1, 3, 0, 0]], (2, 2, 2)),
    ],
)
def test_bad_mapping_true(tableau, conjugacy_class):
    "Test bad_mapping for wrong mapped tableaux, returns True"
    assert ap.bad_mapping(tableau, conjugacy_class)


@pytest.mark.parametrize(
    "partition, conjugacy_class, caracter",
    [
        ((2, 1), (1, 1, 1), 2),
        ((3, 1), (1, 1, 1, 1), 3),
        ((3, 1), (3, 1), 0),
        ((3, 1), (2, 2), -1),
        ((3, 1), (2, 1, 1), 1),
        ((3, 1), (1, 2, 1), 1),
        ((3, 1), (1, 1, 2), 1),
        ((3, 2), (1, 1, 1, 1, 1), 5),
        ((3, 2), (1, 2, 1, 1), 1),
        ((3, 3), (1, 1, 1, 1, 1, 1), 5),
        ((3, 3), (2, 2, 2), -3),
        ((3, 3), (2, 2, 1, 1), 1),
        ((4, 1), (1, 1, 1, 1, 1), 4),
        ((4, 1), (2, 1, 1, 1), 2),
        ((4, 2), (2, 2, 2, 2), 0),
        ((4, 2), (1, 1, 1, 1, 1, 1), 9),
        ((4, 2), (2, 2, 2), 3),
        ((4, 2), (4, 1, 1), -1),
        ((4, 3), (1, 1, 1, 1, 1, 1, 1), 14),
        ((2, 2, 1), (2, 3), -1),
        ((2, 2, 2), (4, 2), -1),
    ],
)
def test_murn_naka(partition, conjugacy_class, caracter):
    "Test murn_naka based on the outputs form weingarten mathematica package"
    assert ap.murn_naka(partition, conjugacy_class) == caracter


@pytest.mark.parametrize(
    "partition, dimension",
    [
        ((1,), 1),
        ((5,), 1),
        ((3, 2), 5),
        ((11, 3), 273),
        ((8, 2, 2), 616),
        ((11, 3, 2), 13860),
        ((7, 3, 2, 2), 28028),
        ((5, 4, 2, 1, 1, 1), 63063),
        ((4, 2, 1, 1, 1, 5), -7007),
    ],
)
def test_sn_dimension(partition, dimension):
    "Test sn_dimension based on the outputs form weingarten mathematica package"
    assert ap.sn_dimension(partition) == dimension


@pytest.mark.parametrize(
    "partition",
    [
        ((1,)),
        ((5,)),
        ((3, 2)),
        ((11, 3)),
        ((8, 2, 2)),
        ((5, 3, 2)),
        ((4, 3, 2, 2)),
        ((4, 3, 2, 1, 1)),
    ],
)
def test_sn_dimension_murn_naka(partition):
    "Reconcil sn_dimension and murn_naka for a class mu of ones"
    mu = sum(partition) * [1]
    assert ap.sn_dimension(partition) == ap.murn_naka(partition, mu)


@pytest.mark.parametrize(
    "partition, dimension",
    [
        ((1,), 17),
        ((5,), 20349),
        ((3, 2), 65892),
        ((11, 3), 7979191740),
        ((8, 2, 2), 2489487616),
        ((7, 3, 2, 2), 98023574880),
        ((5, 4, 2, 1, 1, 1), 86129014608),
        ((6, 5, 4), 168562278720),
        ((11, 3, 2), 405097426800),
    ],
)
def test_ud_dimension(partition, dimension):
    "Test ud_dimension based on the outputs form weingarten mathematica package"
    assert ap.ud_dimension(partition, 17) == dimension


@pytest.mark.parametrize(
    "partition",
    [
        ((3, 2)),
        ((11, 3)),
        ((8, 2, 2)),
        ((7, 3, 2, 2)),
        ((5, 4, 2, 1, 1, 1)),
        ((6, 5, 4)),
    ],
)
def test_ud_dimension_wrong_dimension(partition):
    "ud_dimension returns 0 if the dimension is lower than the number of parts in the partition"
    assert not ap.ud_dimension(partition, len(partition) - 1)


@pytest.mark.parametrize(
    "conjugacy, num, denum",
    [
        ((1,), 1, 7),
        ((2,), -1, 336),
        ((1, 1), 1, 48),
        ((2, 1), -1, 2160),
        ((1, 1, 1), 47, 15120),
        ((3, 1), 19, 846720),
        ((2, 2), 11, 846720),
        ((3, 2), -61, 69854400),
        ((3, 1, 1), 1, 249480),
        ((2, 2, 1), 47, 19958400),
        ((3, 3), 311, 3353011200),
        ((2, 1, 1, 1, 1), -421, 191600640),
        ((1, 1, 1, 1, 1, 1), 82477, 6706022400),
        ((4, 3), -17, 792529920),
    ],
)
def test_weingarten_class(conjugacy, num, denum):
    "Test weingarten_class based on the outputs form weingarten mathematica package"
    assert ap.weingarten_class(conjugacy, 7) == Fraction(num, denum)


@pytest.mark.parametrize(
    "cycle, degree, num, denum",
    [
        (Permutation(0, 1), 2, -1, 336),
        (Permutation(0), 2, 1, 48),
        (Permutation(0, 1, 2), 3, 1, 7560),
        (Permutation(1, 2), 3, -1, 2160),
        (Permutation(2), 3, 47, 15120),
        (Permutation(0, 1, 2), 4, 19, 846720),
        (Permutation(0, 2)(1, 3), 4, 11, 846720),
        (Permutation(2), 4, 403, 846720),
        (Permutation(4, 1), 5, -1739, 139708800),
        (Permutation(2), 5, 1499, 19958400),
        (Permutation(3, 4, 5), 6, 5167, 6706022400),
    ],
)
def test_weingarten_element(cycle, degree, num, denum):
    "Test weingarten_element based on the outputs form weingarten mathematica package"
    assert ap.weingarten_element(cycle, degree, 7) == Fraction(num, denum)


@pytest.mark.parametrize(
    "cycle, degree",
    [
        (Permutation(0, 1), 2),
        (Permutation(0), 2),
        (Permutation(0, 1, 2), 3),
        (Permutation(1, 2), 3),
        (Permutation(2), 3),
        (Permutation(0, 1, 2), 4),
        (Permutation(0, 2)(1, 3), 4),
        (Permutation(2), 4),
        (Permutation(4, 1), 5),
        (Permutation(2), 5),
        (Permutation(3, 4, 5), 6),
    ],
)
def test_weingarten_reconciliation(cycle, degree):
    "Numeric reconciliation of weingarten_class and weingarten_element"
    assert ap.weingarten_element(cycle, degree, 9) == ap.weingarten_class(
        ap.get_class(cycle, degree), 9
    )


@pytest.mark.parametrize(
    "cycle, degree",
    [
        (Permutation(0, 1), 2),
        (Permutation(0), 2),
        (Permutation(0, 1, 2), 3),
        (Permutation(1, 2), 3),
        (Permutation(2), 3),
        (Permutation(0, 1, 2), 4),
        (Permutation(0, 2)(1, 3), 4),
        (Permutation(2), 4),
        (Permutation(4, 1), 5),
        (Permutation(2), 5),
        (Permutation(3, 4, 5), 6),
    ],
)
def test_weingarten_reconciliation_symbolic(cycle, degree):
    "Symbolic reconciliation of weingarten_class and weingarten_element"
    d = Symbol("d")
    assert ap.weingarten_element(cycle, degree, d) == ap.weingarten_class(
        list(ap.get_class(cycle, degree)), d
    )



# Copyright 2024 Polyquantique

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
about tests
"""

import contextlib
import io
import re

import haarpy as ap


def test_about():
    """
    about: Tests if the about string prints correctly.
    """
    f = io.StringIO()
    with contextlib.redirect_stdout(f):
        ap.about()
    out = f.getvalue().strip()

    assert "Python version:" in out
    pl_version_match = re.search(r"Haarpy version:\s+([\S]+)\n", out).group(1)
    assert ap.version() in pl_version_match
    assert "Numpy version" in out
    assert "SymPy version" in out